from house_gen import House
from plot_house_graph import plot
import networkx as nx
import pickle
import logging

NUM_HOUSES = 5

if __name__ == '__main__':
	num_times = 0
	while num_times < NUM_HOUSES:
		try:
			h=House()
			vals = h.gen()
			if vals is not None:
				scene_graph_list = nx.to_dict_of_lists(vals[3])
				meta_data = vals[2]
				img = vals[1]
				storage_dict = {"world": img, "data": meta_data, "scene_graph": scene_graph_list}
				pickle.dump(storage_dict, open(f'/home/isaac/grid_graphs/environments/house_env/houses/house{num_times}.pickle', 'wb'))
				num_times += 1
				# logging.info("Wrote world files successfully")
			else:
				logging.warning("Regenerating...")
		except TimeoutError as e:
			pass

	# storage_dict = pickle.load(open('/home/isaac/grid_graphs/environments/house_env/houses/house0.pickle', 'rb'))
	# scene_graph = nx.Graph(storage_dict["scene_graph"])
	# plot(storage_dict["data"], scene_graph)


from dataclasses import dataclass

@dataclass
class living_room:
	name = "living_room"
	room_associations = {"dining_room": 0.8, "kitchen": 0.7, "utility_closet": 0.2, "bedroom": 0.6}
	furniture_associations = {"couch": 0.9, "tv_set": 0.8, "bookshelf": 0.6, "coffee_table": 0.7, "chair": 0.7}
	object_associations = {"tv_remote": 0.8, "chess_board": 0.2, "magazine": 0.5}
	color = (21, 21, 203) # Red
	existence = 1.0
	duplicate = 0.0

@dataclass
class garage:
	room = "garage"
	room_associations = {"living_room": 0.1, "utility_closet": 0.8, "basement": 0.6}
	furniture_associations = {"fridge": 0.2, "shelf": 0.9}
	object_associations = {"wrench": 0.8, "drill": 0.8}
	color = (96, 96, 96) # Grey
	existence = 0.4
	duplicate = 0.0

@dataclass
class kitchen:
	room = "kitchen"
	room_associations = {"dining_room": 0.9, "living_room": 0.7, "pantry": 0.6, "bathroom": 0.7}
	furniture_associations = {"fridge": 1.0, "cupboard": 0.9, "counter_top": 0.9, "sink": 1.0, "dish_washer": 0.6}
	object_associations = {"fork": 1.0, "plate": 1.0, "soap": 1.0, "spices": 1.0, "pot": 1.0, "cup": 1.0}
	color = (0, 204, 102) # Green
	existence = 1.0
	duplicate = 0.0

@dataclass
class pantry:
	room = "pantry"
	room_associations = {"kitchen": 0.9}
	furniture_associations = {"shelf": 1.0}
	object_associations = {"rice": 0.6, "molasses": 0.4, "cereal": 0.8, "flour": 0.8, "spices": 0.9, "dog_food": 0.4}
	color = (255, 102, 178) # Lavendar
	existence = 0.9
	duplicate = 0.0

@dataclass
class bedroom:
	room = "bedroom"
	room_associations = {"dining_room": 0.7, "living_room": 0.7, "bathroom": 0.7}
	furniture_associations = {"bookshelf": 0.8, "bed": 1.0, "desk": 0.8, "nightstand": 0.9}
	object_associations = {"lamp": 1.0, "laptop": 0.7, "book": 0.8, "suit": 0.6}
	color = (204, 102, 0) # Dark blue
	existence = 1.0
	duplicate = 0.8

@dataclass
class bathroom:
	room = "bathroom"
	room_associations = {"dining_room": 0.6, "kitchen": 0.7, "bedroom": 0.7, "living_room": 0.7}
	furniture_associations = {"sink": 1.0, "cupboard": 0.9, "toilet": 1.0, "bathtub": 0.7, "shelf": 0.6}
	object_associations = {"toothbrush": 1.0, "toilet_paper": 1.0, "medicine": 0.8, "mouthwash": 0.8}
	color = (254, 255, 243) # Eggshell
	existence = 1.0
	duplicate = 0.5

@dataclass
class basement:
	room = "basement"
	room_associations = {"bathroom": 0.4, "utility_closet": 0.6}
	furniture_associations = {"table": 0.9, "shelf": 0.6, "washing_machine": 0.6}
	object_associations = {"laundry_detergent": 0.9, "skis": 0.4, "vacuum_cleaner": 0.8}
	color = (105, 114, 60) # Turqoise grey
	existence = 0.3
	duplicate = 0.0

@dataclass
class utility_closet:
	room = "utility_closet"
	room_associations = {"basement": 0.5, "kitchen": 0.5}
	furniture_associations = {"shelf": 1.0}
	object_associations = {"mop": 0.9, "broom": 0.9, "dustpan": 0.9, "vacuum_cleaner": 0.5, "bleach": 0.6}
	color = (0, 51, 102) # Brown
	existence = 0.4
	duplicate = 0.1

@dataclass
class dining_room:
	room = "dining_room"
	room_associations = {"living_room": 0.9, "kitchen": 0.9, "bedroom":0.6}
	furniture_associations = {"table": 1.0, "chair": 1.0}
	object_associations = {"plate": 0.6, "cup": 0.7, "fork": 0.4, "vase": 0.7}
	color = (255, 255, 0) # Cyan
	existence = 1.0
	duplicate = 0.0


from house_gen import House
import plotly.graph_objects as go

def plot(meta_data, graph):

    coords = {}
    colors = {}
    for node in meta_data.keys():
        x_y = (meta_data[node]["centroid"][1], meta_data[node]["centroid"][0]) 
        if meta_data[node]["class"] == "building":
            coords[meta_data[node]["name"]] = x_y + (500,)
            colors[meta_data[node]["name"]] = 1
        elif meta_data[node]["class"] == "room":
            coords[meta_data[node]["name"]] = x_y + (400,)
            colors[meta_data[node]["name"]] = 2
        elif meta_data[node]["class"] == "furniture":
            coords[meta_data[node]["name"]] = x_y + (100,)
            colors[meta_data[node]["name"]] = 3

    # Extract the coordinates of the nodes for plotting
    X, Y, Z = zip(*coords.values())

    # Create a Scatter plot for nodes
    node_trace = go.Scatter3d(
        x=X,
        y=Y,
        z=Z,
        mode='markers+text',
        text=[f"{node}" for node in coords.keys()],  # Labels for all nodes
        textposition='bottom center',  # Position of the labels
        textfont=dict(size=25),
        marker=dict(size=12)
    )

    node_colors = [colors[node] for node in colors.keys()]


    # Create an edge trace for lines connecting the nodes
    edge_x = []
    edge_y = []
    edge_z = []

    for edge in graph.edges():
        # room_name = meta_data[]
        node1_idx, node2_idx = edge
        x0, y0, z0 = coords[node1_idx]
        x1, y1, z1 = coords[node2_idx]
        edge_x.extend([x0, x1, None])
        edge_y.extend([y0, y1, None])
        edge_z.extend([z0, z1, None])

    edge_trace = go.Scatter3d(
        x=edge_x,
        y=edge_y,
        z=edge_z,
        mode='lines',
        line=dict(width=3.0, color='black')
        )

    axis = dict(showbackground=False,
            showline=False,
            zeroline=False,
            showgrid=False,
            showticklabels=False,
            title=''
            )


    layout = go.Layout(title="Scene-graph", showlegend=False, width=1000, height=1000, scene=dict(
             xaxis=dict(axis),
             yaxis=dict(axis),
             zaxis=dict(axis),
            ),
     margin=dict(
        t=100
    ), )

    fig = go.Figure(data=[edge_trace, node_trace], layout=layout)
    fig.update_traces(marker=dict(color=node_colors, colorscale='Rainbow', cmin=min(node_colors), cmax=max(node_colors)))
    fig.show()

# plot(meta_data, graph)


from . import house_rooms, house_furniture
import networkx as nx
import random as rd
import numpy as np
import math
import cv2
from copy import deepcopy
import time
from .house_gen_timeout import timeout

SCALE_X = 700
SCALE_Y = 700
NUM_REGIONS = 6
BLUR_KERNEL = 37
DOORWAY_WIDTH = 70
TIMEOUT = 15

class House:
	def __init__(self) -> None:
		self.__graph = nx.Graph()

	@timeout(15)
	def gen(self):
		return self.__gen_graph()


	def __gen_graph(self):

		def find_distance(p1, p2):
			"""
			Returns L2 distance between two points (x, y)
			"""
			return np.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)

		inverted_image, contours, centroids = self.__gen_grid()

		height, width = inverted_image.shape
		color_image = np.zeros((height, width, 3), dtype=np.uint8)

		color_image[:, :, 0] = inverted_image  # Blue channel
		color_image[:, :, 1] = inverted_image  # Green channel
		color_image[:, :, 2] = inverted_image  # Red channel

		init_rooms_dict = {}
		for i in range(1, len(contours) + 1):
			init_rooms_dict[i] = {"contour": contours[i - 1], "centroid": centroids[i - 1]} # last index represents shared-edges
		
		def is_point_near_edge(x, y, distance_threshold):
			"""
			Calculate distances from the point to a specified border (for edge of matrix) and return True if so, False otherwise.
			"""
			distance_left = x
			distance_right = SCALE_X - x - 1
			distance_top = y
			distance_bottom = SCALE_Y - y - 1
			
			# Check if the point is within the distance threshold from any edge
			if (distance_left <= distance_threshold or
				distance_right <= distance_threshold or
				distance_top <= distance_threshold or
				distance_bottom <= distance_threshold):
				return True
			else:
				return False
				
		def find_midpoint(point1, point2):
			x_mid = (point1[0] + point2[0]) / 2
			y_mid = (point1[1] + point2[1]) / 2
			return (x_mid, y_mid)

		def find_closest_points(points):
			closest_points = {}
			for i in range(len(points)):
				current_point = tuple(points[i])
				min_distance = float('inf')
				closest_point = None

				for j in range(len(points)):
					if i != j:
						dist = find_distance(current_point, points[j])
						if dist < min_distance:
							min_distance = dist
							closest_point = tuple(points[j])

				closest_points[current_point] = closest_point

			return closest_points
		
		midpoints = []
		connected_rooms = set()
		midpoints_to_rooms  = {}

		for room in init_rooms_dict.keys():
			connected_point_sets = []

			# Iterate through the points in the contour
			for i in range(1, len(init_rooms_dict[room]["contour"])):
				
				point = init_rooms_dict[room]["contour"][i][0]
				prev_point = init_rooms_dict[room]["contour"][i - 1][0]

				# Calculate the distance between the current point and the previous point
				dist = find_distance(point, prev_point)

				both_edge = False
				if is_point_near_edge(point[0], point[1], 50) and is_point_near_edge(prev_point[0], prev_point[1], 50):
					both_edge = True

				# Check if the distance is greater than the threshold
				if dist > (DOORWAY_WIDTH * 1.9) and not both_edge: #:
					# If the distance is greater than the threshold, start a new segment
					connected_point_sets.append([point, prev_point])

			point = init_rooms_dict[room]["contour"][-1][0]
			prev_point = init_rooms_dict[room]["contour"][0][0]
			dist = find_distance(point, prev_point)
			both_edge = False
			
			if is_point_near_edge(point[0], point[1], 50) and is_point_near_edge(prev_point[0], prev_point[1], 50):
				both_edge = True
			if dist > (DOORWAY_WIDTH * 2) and not both_edge:
				connected_point_sets.append([point, prev_point])

			for p_set in connected_point_sets:
				midpoint = np.int_(find_midpoint(p_set[0], p_set[1]))
				midpoints.append(midpoint)
				midpoints_to_rooms[tuple(midpoint)] = room
		
		closest_points = find_closest_points(midpoints)

		rooms_to_doorways = {}

		# Create list of edges based on the closest points between all the edge-midpoints.
		# These will inform the location of doorways between rooms.
		for point, closest_point in closest_points.items():
			if closest_points[closest_point] == point:
				doorway = np.int_(find_midpoint(point, closest_point))
				closest_room = midpoints_to_rooms[closest_point]
				this_room = midpoints_to_rooms[point]
				if all(color_image[doorway[1], doorway[0]] == (0, 0, 0)) and (closest_room, this_room) not in connected_rooms:
					# doorways.append(doorway)
					for room_p in (closest_room, this_room):
						if room_p not in rooms_to_doorways.keys():
							rooms_to_doorways[room_p] = (doorway,)
						else:
							rooms_to_doorways[room_p] += (doorway,)
					
					connected_rooms.add((midpoints_to_rooms[point], midpoints_to_rooms[closest_point]))
					cv2.rectangle(color_image, (int(doorway[0] - BLUR_KERNEL / 1.5), int(doorway[1] + BLUR_KERNEL / 1.5)), (int(doorway[0] + BLUR_KERNEL / 1.5), int(doorway[1] - BLUR_KERNEL / 1.5)), (255, 255, 255), thickness=-1)

		graph_obj_dict = {}

		# Construct updated dictionary containing only valid rooms, i.e. those with edges
		for connection in connected_rooms:
			room1, room2 = connection
			for room in [room1, room2]:
				if room not in graph_obj_dict.keys():
					attributes = init_rooms_dict[room]
					doorways = rooms_to_doorways[room]
					edges = (room2,) if room == room1 else (room1,)
					attributes["edges"] = edges
					attributes["doorways"] = doorways
					attributes["class"] = "room"
					graph_obj_dict[room] = attributes
				else:
					graph_obj_dict[room]["edges"] += (room2,) if room == room1 else (room1,)

		print("Num Doorways: ", str(len(connected_rooms)))
		print("Num Rooms: ", str(len(graph_obj_dict)))

		max_room_id, _ = max(graph_obj_dict.items(), key=lambda item: cv2.contourArea(item[1]["contour"]))

		graph_obj_dict[max_room_id]["name"] = "living_room0"

		def dfs(room_dict, start_node):
			"""
			Populates the room layer of the scene-graph procedurally, given a dictionary of edges between the rooms
			and the starting node to generate from (living room is probably best).
			"""
			visited = set()
			queue = [start_node]
			visited.add(start_node)
			occurances_dict = {room_dict[start_node]["name"]: 1}
			self.__graph.add_edge("house", room_dict[start_node]["name"])

			while queue:
				node = queue.pop(0)
				neighbors = room_dict[node]["edges"]
				adj_rooms, adj_room_scores = zip(*getattr(house_rooms, room_dict[node]["name"][:-1]).room_associations.items())

				for neighbor in neighbors:
					if neighbor not in visited:
						room_name = rd.choices(adj_rooms, adj_room_scores)[0] + "0"

						duplicate = getattr(house_rooms, room_name[:-1]).duplicate
						again = 0
						while(self.__graph.has_edge("house", room_name) and again == 0):
							duplicate = getattr(house_rooms, room_name[:-1]).duplicate
							if duplicate > 0.01:
								again = np.random.binomial(1, duplicate ** occurances_dict[room_name[:-1]])
								if again == 1:
									room_name = room_name[:-1] + str(occurances_dict[room_name[:-1]])
							else:
								room_name = rd.choices(adj_rooms, adj_room_scores)[0]
								room_name += "0"

						self.__graph.add_edge("house", room_name)
						self.__graph.add_edge(room_dict[node]["name"], room_name)
						room_dict[neighbor]["name"] = room_name
						
						if room_name[:-1] not in  occurances_dict.keys():
							occurances_dict[room_name[:-1]] = 1
						else:
							occurances_dict[room_name[:-1]] += 1

						visited.add(neighbor)
						queue.append(neighbor)
					else:
						self.__graph.add_edge(room_dict[node]["name"], room_dict[neighbor]["name"])

		dfs(graph_obj_dict, max_room_id)

		return_img = deepcopy(color_image)

		graph_obj_dict[0] = {"name": "house", "centroid": (int(SCALE_X / 2), int(SCALE_Y / 2)), "contour":(), "class":"building"}

		house_edges = tuple([i for i in range(1, len(graph_obj_dict) + 1)])

		graph_obj_dict[0]["edges"] = house_edges

		def rand_contour_point(contour, max_distance_from_edge, min_distance_from_edge, doorways, f_pieces):
			while True:
				# Get the bounding rectangle of the contour
				x, y, w, h = cv2.boundingRect(contour)
				
				# Generate a random point within the bounding rectangle
				random_point = (x + np.random.randint(0, w), y + np.random.randint(0, h))
				
				# Check if the random point is within the contour and has the specified distance from its edges
				p_dist = cv2.pointPolygonTest(contour, random_point, True)
				if p_dist < max_distance_from_edge and p_dist > min_distance_from_edge:
					near_doorway = False
					for doorway in doorways:
						if find_distance(random_point, doorway) < (DOORWAY_WIDTH * 2):
							near_doorway = True
							break
					
					near_furniture = False
					for f in f_pieces:
						if find_distance(random_point, f) < (DOORWAY_WIDTH):
							near_furniture = True
							break

					if not near_doorway and not near_furniture:
						return random_point

		max_id = max(graph_obj_dict.keys())
		rooms = list(graph_obj_dict.keys())
		furniture_nums = {}

		for room_id in rooms:
			if room_id != 0:
				f_pieces = tuple()
				room_f_dict = getattr(house_rooms, graph_obj_dict[room_id]["name"][:-1]).furniture_associations.keys()
				num_pieces = len(room_f_dict)
				for f_id, f in enumerate(room_f_dict, max_id + 1):
					if f not in furniture_nums.keys():
						furniture_nums[f] = 0
					else:
						furniture_nums[f] += 1
					f_name = f + str(furniture_nums[f])
					contour = graph_obj_dict[room_id]["contour"]
					point = rand_contour_point(contour, 30, 5, graph_obj_dict[room_id]["doorways"], f_pieces)
					f_pieces += (point,)
					graph_obj_dict[f_id] = {"name":f_name, "centroid":point, "edges": (room_id), "contour":(), "class":"furniture"}
					graph_obj_dict[room_id]["edges"] += (f_id,)
					self.__graph.add_edge(graph_obj_dict[room_id]["name"], f_name)
					cv2.circle(color_image, point, 5, (255, 0, 0), -1)
				max_id += num_pieces


		for obj in graph_obj_dict:
			text = graph_obj_dict[obj]["name"]
			font = cv2.FONT_HERSHEY_SIMPLEX
			font_scale = 0.5
			font_color = (0, 0, 0)  # black color in BGR format
			line_thickness = 2
			text_size, _ = cv2.getTextSize(text, font, font_scale, line_thickness)
			position = graph_obj_dict[obj]["centroid"]
			text_x = position[0] - text_size[0] // 2
			text_y = position[1] + text_size[1] // 2
			cv2.putText(color_image, text, (text_x, text_y), font, font_scale, font_color, line_thickness)

		return return_img, color_image, graph_obj_dict, self.__graph


	def __gen_grid(self):
		'''
		Generates a randomly generated grid, where the integers filling the grid represent different types
		of rooms.
		'''
		floor_matrix = np.zeros((SCALE_Y, SCALE_X))

		def calculate_distance(row1, col1, row2, col2):
				return math.sqrt((row1 - row2)**2 + (col1 - col2)**2)

		def place_integers(matrix, num_integers, min_distance, max_distance):
				# Randomly place the first integer

			new_pos = {}
			row = rd.randint(0, len(matrix) - 1)
			col = rd.randint(0, len(matrix[0]) - 1)
			matrix[row][col] = 1

			new_pos[1] = (row, col)

			# Place the remaining integers
			for i in range(2, num_integers + 1):
				placed = False

				while not placed:
					# Randomly select a cell
					row = rd.randint(0, len(matrix) - 1)
					col = rd.randint(0, len(matrix[0]) - 1)

					# Check if the cell satisfies the distance constraint with existing integers
					valid_placement = True
					for r in range(len(matrix)):
						for c in range(len(matrix[0])):
							if matrix[r][c] != 0:
								distance = calculate_distance(row, col, r, c)
								if distance < min_distance or distance > max_distance:
									valid_placement = False
									break
						
						if not valid_placement:
							break

					if valid_placement:
						matrix[row][col] = i
						new_pos[i] = (row, col)
						placed = True

			return matrix, new_pos


		def generate_square_coordinates(distance):
			coordinates = []

			top_left = (-distance, distance)
			top_right = (distance, distance)
			bottom_left = (-distance, -distance)
			bottom_right = (distance, -distance)

			# Generate coordinates along each side of the square
			for x in range(top_left[0], top_right[0] + 1):
				coordinates.append((x, top_left[1]))

			for y in range(top_right[1], bottom_right[1] - 1, -1):
				coordinates.append((top_right[0], y))

			for x in range(bottom_right[0], bottom_left[0] - 1, -1):
				coordinates.append((x, bottom_right[1]))

			for y in range(bottom_left[1], top_left[1] + 1):
				coordinates.append((bottom_left[0], y))

			return coordinates


		floor_matrix, new_pos = place_integers(floor_matrix, NUM_REGIONS, int(0.25*SCALE_X), int(0.75*SCALE_X))
		num_zeros = (SCALE_X * SCALE_Y) - len(new_pos)
		distance = 1

		while num_zeros > 0:
			for idx, room in enumerate(new_pos):
				directions = generate_square_coordinates(distance)
				for dy, dx in directions:
					new_row = new_pos[room][0] + dy
					new_col = new_pos[room][1] + dx
					
					if 0 <= new_row < SCALE_Y and 0 <= new_col < SCALE_X:
						if floor_matrix[new_row][new_col] == 0:
							floor_matrix[new_row][new_col] = idx + 1
							num_zeros -= 1
			distance += 1

		def separate_polygons(matrix):
			integers = np.unique(floor_matrix)
			img_list = []
			cpy = deepcopy(matrix)
			for i in integers:
				cpy[cpy != i] = 0
				img_list.append(cpy)
				cpy=deepcopy(matrix)

			return img_list

		img_list = separate_polygons(floor_matrix)		

		def convert_to_image(grid, color_mapping):
				num_rows = len(grid)
				num_cols = len(grid[0])

				# Create an empty image array
				image = np.zeros((num_rows, num_cols, 3), dtype=np.uint8)

				# Iterate over the grid and assign colors based on the integer values
				for row in range(num_rows):
						for col in range(num_cols):
								integer = grid[row][col]
								color = color_mapping.get(integer, (0, 0, 0))  # Default to black if integer not in color mapping
								image[row, col] = color

				return image

		# Define a color mapping for integers
		color_mapping = {
				1: (0, 0, 255),   # Blue
				2: (0, 255, 0),   # Green
				3: (255, 0, 0),   # Red
				4: (0, 255, 255),  # Cyan
				5: (120, 120, 120),
				6: (65, 180, 15),
				7: (92, 23, 200),
				8: (200, 54, 10)
		}

		def convert_to_rectangle(image):
				# Convert the image to grayscale
				gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
				
				# # Find contours of polygons in the grayscale image
				contours, _ = cv2.findContours(gray, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
				
				# Create a new image to draw the rectangles
				result = np.zeros_like(gray)
				
				for contour in contours:
						# Find the bounding rectangle for each contour
						x, y, width, height = cv2.boundingRect(contour)
						
						# Fill the rectangle with the mean value of the polygon region
						result[y:y+height, x:x+width] = np.mean(gray[contour[:, :, 1], contour[:, :, 0]])
				
				return result
		
		start = time.time()
		size_img_dict = {}
		rects_list = []
		
		for idx, i in enumerate(img_list):
				max_size = 0
				cpy = convert_to_image(i, color_mapping)
				val = 0
				for x in range(len(i)):
					for y in range(len(i[0])):
						if i[x][y] != 0:
							val = i[x][y]
							break
				
				rect = convert_to_rectangle(cpy)
				black_image = np.zeros((len(i), len(i[0]), 3), dtype=np.uint8)
				for j in range(len(i)):
					for k in range(len(i[0])):
						if rect[j][k] != 0:
							black_image[j][k] = color_mapping[val]
							max_size += 1

				size_img_dict[idx] = max_size
				rects_list.append(black_image)
		
		end = time.time()
		print("Generation time: ", str(end-start))

		sorted_keys = sorted(size_img_dict, key=lambda k: size_img_dict[k], reverse=True)

		black_image = np.zeros((len(img_list[0]), len(img_list[0][0]), 3), dtype=np.uint8)
		for idx in sorted_keys:
			curr_img = rects_list[idx]
			for i in range(len(curr_img)):
				for j in range(len(curr_img[0])):
					if curr_img[i][j].any():
						black_image[i][j] = curr_img[i][j]

		gray = cv2.cvtColor(black_image, cv2.COLOR_BGR2GRAY)

		# Apply Canny edge detection to identify edges
		edges = cv2.Canny(gray, 100, 200)

		# Apply Gaussian blur to smooth the edges (which will represent the walls between rooms)
		blurred_edges = cv2.GaussianBlur(edges, (BLUR_KERNEL, BLUR_KERNEL), 0)  # You can adjust the kernel size for stronger or weaker smoothing

		# Set white pixels over the edges
		for i in range(len(blurred_edges)):
				for j in range(len(blurred_edges[0])):
						if blurred_edges[i, j] > 0:  # Check if the pixel value is greater than 0 (indicating an edge)
								blurred_edges[i, j] = 255  # Set the pixel to white
					
		border_size = (50, 50)


		# Add a white border with the same size as the Gaussian blur kernel
		bordered_image = cv2.copyMakeBorder(blurred_edges, border_size[0], border_size[0], border_size[1], border_size[1], cv2.BORDER_CONSTANT, value=(255, 255, 255))

		inverted_image = cv2.bitwise_not(bordered_image)

		contours, _ = cv2.findContours(inverted_image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

		# Initialize a list to store the centroids of each polygon
		centroids = []

		# Loop through each identified contour (polygon)
		for contour in contours:
				# Calculate the moments of the contour
				M = cv2.moments(contour)

				# Calculate the x and y coordinates of the centroid
				centroid_x = int(M["m10"] / M["m00"])
				centroid_y = int(M["m01"] / M["m00"])

				# Append the centroid coordinates to the list
				centroids.append((centroid_x, centroid_y))

		return inverted_image, contours, centroids


from dataclasses import dataclass

@dataclass
class couch:
  name = "couch"
  object_associations = {}
  duplicate = 0.3

@dataclass
class tv_set:
  name = "tv_set"
  object_associations = {}
  duplicate = 0.0

@dataclass
class bookshelf:
	name = "bookshelf"
	object_associations = {}
	duplicate = 0.1

@dataclass
class coffee_table:
	name = "coffee_table"
	object_associations = {}
	duplicate = 0.1
        
@dataclass
class fridge:
	name = "fridge"
	object_associations = {}
	duplicate = 0.2
        
@dataclass
class chair:
	name = "chair"
	object_associations = {}
	duplicate = 0.8

@dataclass
class cupboard:
	name = "cupboard"
	object_associations = {}
	duplicate = 0.9

@dataclass
class counter_top:
	name = "counter_top"
	object_associations = {}
	duplicate = 0.5

@dataclass
class dish_washer:
	name = "dish_washer"
	object_associations = {}
	duplicate = 0.0

@dataclass
class bed:
  name = "bed"
  object_associations = {}
  duplicate = 0.2
  
@dataclass
class desk:
  name = "desk"
  object_associations = {}
  duplicate = 0.1
  
@dataclass
class nightstand:
  name = "nightstand"
  object_associations = {}
  duplicate = 0.2
  
@dataclass
class toilet:
  name = "toilet"
  object_associations = {}
  duplicate = 0.0
  
@dataclass
class bathtub:
  name = "bathtub"
  object_associations = {}
  duplicate = 0.0
  
@dataclass
class shelf:
  name = "shelf"
  object_associations = {}
  duplicate = 0.7
  
@dataclass
class washing_machine:
  name = "washing_machine"
  object_associations = {}
  duplicate = 0.0
  
@dataclass
class table:
  name = "table"
  object_associations = {}
  duplicate = 0.0


import signal

class TimeoutError(Exception):
	pass

def timeout(seconds):
	def decorator(func):
		def _handle_timeout(signum, frame):
			raise TimeoutError("Environment generation failed.")

		def wrapper(*args, **kwargs):
			# Set the signal handler for SIGALRM
			signal.signal(signal.SIGALRM, _handle_timeout)
			# Set the alarm to trigger after the specified timeout
			signal.alarm(seconds)
			try:
				# Call the original function
				result = func(*args, **kwargs)
			except TimeoutError as e:
				print("Timeout occurred:", e)
				result = None  # Set a default value or None as desired
			finally:
				# Reset the alarm
				signal.alarm(0)
				return result
		return wrapper
	return decorator
